﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Cursed.gob.mob;
using Microsoft.Xna.Framework;
using System.Windows;
using Cursed.gob;
using Cursed.gob.tile;

namespace Cursed.col
{
    

    class QTNode
    {
        const int max = 5;
        const float minSize = 1;

        Rect bounds;
        QTNode[] subtrees = { null, null, null, null };
        ICollection<Gob> gobs = new HashSet<Gob>();
        bool hasChildren;

        public QTNode(Rect bounds)
        {
            this.bounds = bounds;
        }



        public void Add(Gob gob)
        {
            if (!hasChildren)
            {
                gobs.Add(gob);
                if (gobs.Count == max && (bounds.Width > minSize && bounds.Height > minSize))
                {
                    hasChildren = true;
                    double w = bounds.Width * .5;
                    double h = bounds.Height * .5;
                    double x = bounds.X;
                    double y = bounds.Y;
                    double cx = bounds.X + w;
                    double cy = bounds.Y + h;

                    subtrees[0] = new QTNode(new Rect(x, y, w, h));
                    subtrees[1] = new QTNode(new Rect(x, y+h, w, h));
                    subtrees[2] = new QTNode(new Rect(x+h, y, w, h));
                    subtrees[3] = new QTNode(new Rect(x+h, y+h, w, h));

                    foreach (Gob g in gobs)
                        addSub(g);
                }
            }
            else
            {
                addSub(gob);
            }
        }

        void addSub(Gob gob)
        {
            Rect bounds = gob.Bounds;
            foreach (QTNode qtNode in subtrees)
            {
                if (qtNode.bounds.IntersectsWith(bounds))
                    qtNode.Add(gob);
            }
        }

        public void Collect(ICollection<Gob> pool, Rect bounds)
        {
            if (!hasChildren)
            {
                foreach (Gob gob in gobs)
                {
                    if (bounds.IntersectsWith(gob.Bounds))
                        pool.Add(gob);
                }
            }
            else
            {
                foreach (QTNode qtNode in subtrees)
                {
                    if (qtNode.bounds.IntersectsWith(bounds))
                        qtNode.Collect(pool, bounds);
                }
            }
        }
    }

    public class QTCollider : ICollider
    {
        ICollection<Mob> mobs;
        QTNode root;

        ICollection<Tile> tiles;
        ICollection<Tile>[,] tileGrid;
        int minTileX;
        int maxTileX;
        int spanTileX;
        int minTileY;
        int maxTileY;
        int spanTileY;

        public QTCollider()
        {
            mobs = new HashSet<Mob>();
            root = null;
            tiles = new List<Tile>();

        }

        void repackTiles()
        {
            minTileX = int.MaxValue;
            maxTileX = -1 * int.MaxValue;
            minTileY = int.MaxValue;
            maxTileY = -1 * int.MaxValue;
            foreach (Tile tile in tiles)
            {
                if (tile.X < minTileX) minTileX = tile.X;
                if (tile.Y < minTileY) minTileY = tile.Y;
                if (tile.X > maxTileX) maxTileX = tile.X;
                if (tile.Y > maxTileY) maxTileY = tile.Y;
            }
            spanTileX = maxTileX - minTileX + 1;
            spanTileY = maxTileY - minTileY + 1;

            if (spanTileX > 0)
            {
                tileGrid = new ICollection<Tile>[spanTileX, spanTileY];
                for (int x = 0; x < spanTileX; x++)
                    for (int y = 0; y < spanTileY; y++)
                    {
                        tileGrid[x, y] = new List<Tile>();
                    }
                foreach (Tile tile in tiles)
                {
                    tileGrid[tile.X - minTileX, tile.Y - minTileY].Add(tile);
                }
            }
        }
        ICollection<Tile> tilesAt(int x, int y)
        {
            return tileGrid[x - minTileX, y - minTileY];
        }

        #region ICollider Members

        public void handleCollisions()
        {
            Rect bounds = new Rect(0, 0, 1, 1);
            foreach (Mob gob in mobs)
            {
                bounds.Union(gob.Bounds);
            }

            root = new QTNode(bounds);
            foreach (Mob gob in mobs)
            {
                root.Add(gob);
            }

            foreach (Mob mob in mobs)
            {
                mob.clearTouch();

                HashSet<Gob> boundTouchedGobs = new HashSet<Gob>();
                root.Collect(boundTouchedGobs, mob.Bounds);
                foreach (Gob o in boundTouchedGobs)
                {
                    if (o == mob)
                        continue;
                    if (mob.touches(o))
                        mob.touch(o);
                }
            }

            foreach (Tile tile in tiles)
                tile.clearTouch();

            foreach (Mob mob in mobs)
            {
                float w = (float)mob.Bounds.Width * .5f + .5f;
                float h = (float)mob.Bounds.Height * .5f + .5f;
                Vector2 c = mob.Loc;
                int minX = (int)(c.X - w);
                int minY = (int)(c.Y - h);
                int maxX = (int)(c.X + w);
                int maxY = (int)(c.Y + h);

                if (minX < minTileX)
                    minX = minTileX;
                if (minY < minTileY)
                    minY = minTileY;
                if (maxX > maxTileX)
                    maxX = maxTileX;
                if (maxY > maxTileY)
                    maxY = maxTileY;

                for (int x = minX; x <= maxX; x++)
                {
                    for (int y = minY; y <= maxY; y++)
                    {

                        ICollection<Tile> tilesHere = tilesAt(x, y);

                        if (tilesHere.Count > 0)
                        {
                            IEnumerator<Tile> tileEnumerator = tilesHere.GetEnumerator();
                            tileEnumerator.MoveNext();
                            Tile top = tileEnumerator.Current;

                            if (mob.touches(top))
                            {
                                foreach (Tile tile in tilesHere)
                                {
                                    mob.touch(tile);
                                    tile.touch(mob);
                                }
                            }
                        }
                    }
                }
            }
        }

        public void AddMob(Cursed.gob.mob.Mob mob)
        {
            mobs.Add(mob);
        }

        public void RemoveMob(Cursed.gob.mob.Mob mob)
        {
            mobs.Remove(mob);
        }

        public void AddTile(Cursed.gob.tile.Tile tile)
        {
            tiles.Add(tile);
            if (tile.TileInfo.X > maxTileX || tile.TileInfo.Y > maxTileY ||
                tile.TileInfo.X < minTileX || tile.TileInfo.Y < minTileY)
            {
                repackTiles();
            }
            else
            {
                tileGrid[tile.X - minTileX, tile.Y - minTileY].Add(tile);
            }
        }

        public void RemoveTile(Cursed.gob.tile.Tile tile)
        {
            tiles.Remove(tile);
            tileGrid[tile.X - minTileX, tile.Y - minTileY].Remove(tile);
        }

        public ICollection<Cursed.gob.mob.Mob> MobsInBounds(System.Windows.Rect bounds)
        {
            HashSet<Gob> boundTouchedGobs = new HashSet<Gob>();
            root.Collect(boundTouchedGobs, bounds);

            HashSet<Mob> mobs = new HashSet<Mob>();
            foreach (Gob gob in boundTouchedGobs)
            {
                if (gob is Mob)
                    mobs.Add((Mob)gob);
            }
            return mobs;
        }

        #endregion
    }
}
