﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Cursed.gob.mob;
using Cursed.gob.tile;
using System.Windows;
using Microsoft.Xna.Framework;
using Cursed.gob.mob.kob;

namespace Cursed.col
{
    public class TileCollider : ICollider
    {
        CursedGame game;

        ICollection<Mob> mobs;
        ICollection<Tile> tiles;

        ICollection<Tile>[,] tileGrid;
        ICollection<Mob>[,] mobGrid;
        ICollection<Tile> emptyTiles;
        ICollection<Mob> emptyMobs;


        int minTileX;
        int maxTileX;
        int spanTileX;
        int minTileY;
        int maxTileY;
        int spanTileY;

        public TileCollider(CursedGame game)
        {
            this.game = game;

            mobs = new List<Mob>();
            tiles = new List<Tile>();
            emptyTiles = new List<Tile>();
            emptyMobs = new List<Mob>();
            repackTiles();
        }

        void repackTiles()
        {
            minTileX = int.MaxValue;
            maxTileX = -1 * int.MaxValue;
            minTileY = int.MaxValue;
            maxTileY = -1 * int.MaxValue;
            foreach (Tile tile in tiles)
            {
                if (tile.X < minTileX) minTileX = tile.X;
                if (tile.Y < minTileY) minTileY = tile.Y;
                if (tile.X > maxTileX) maxTileX = tile.X;
                if (tile.Y > maxTileY) maxTileY = tile.Y;
            }
            spanTileX = maxTileX - minTileX + 1;
            spanTileY = maxTileY - minTileY + 1;

            tileGrid = new ICollection<Tile>[spanTileX, spanTileY];
            mobGrid = new ICollection<Mob>[spanTileX, spanTileY];
            for (int x = 0; x < spanTileX; x++)
                for (int y = 0; y < spanTileY; y++)
                {
                    tileGrid[x, y] = new List<Tile>();
                    mobGrid[x, y] = new List<Mob>();
                }
            foreach (Tile tile in tiles)
            {
                tileGrid[tile.X - minTileX, tile.Y - minTileY].Add(tile);
            }

        }

        public void AddMob(Mob mob)
        {
            mobs.Add(mob);
        }

        public void RemoveMob(Mob mob)
        {
            mobs.Remove(mob);
        }

        public void AddTile(Tile tile)
        {
            tiles.Add(tile);
            if (tile.TileInfo.X > maxTileX || tile.TileInfo.Y > maxTileY ||
                tile.TileInfo.X < minTileX || tile.TileInfo.Y < minTileY)
            {
                repackTiles();
            }
            else
            {
                tileGrid[tile.X - minTileX, tile.Y - minTileY].Add(tile);
            }
        }

        public void RemoveTile(Tile tile)
        {
            tiles.Remove(tile);
            tileGrid[tile.X - minTileX, tile.Y - minTileY].Remove(tile);
        }

        ICollection<Tile> tilesAt(int x, int y)
        {
            return tileGrid[x - minTileX, y - minTileY];
        }

        ICollection<Mob> mobsAt(int x, int y)
        {
            return mobGrid[x - minTileX, y - minTileY];
        }

        public void handleCollisions()
        {

            foreach (Mob mob in mobs)
                mob.clearTouch();
            foreach (Tile tile in tiles)
                tile.clearTouch();

            foreach (ICollection<Mob> mobList in mobGrid)
            {
                if (mobList != null)
                    mobList.Clear();
            }

            game.Profiler.Start("handleCollisions(Tiles)");
            foreach (Mob mob1 in mobs)
            {
                float w = (float)mob1.Bounds.Width * .5f + .5f;
                float h = (float)mob1.Bounds.Height * .5f + .5f;
                Vector2 c = mob1.Loc;
                int minX = (int)(c.X - w);
                int minY = (int)(c.Y - h);
                int maxX = (int)(c.X + w);
                int maxY = (int)(c.Y + h);

                if (minX < minTileX)
                    minX = minTileX;
                if (minY < minTileY)
                    minY = minTileY;
                if (maxX > maxTileX)
                    maxX = maxTileX;
                if (maxY > maxTileY)
                    maxY = maxTileY;

                game.Profiler.Start("handleCollisions(Tiles,Mob)");
                for (int x = minX; x <= maxX; x++)
                {
                    for (int y = minY; y <= maxY; y++)
                    {

                        ICollection<Tile> tilesHere = tilesAt(x, y);
                        mobsAt(x, y).Add(mob1);

                        if (tilesHere.Count > 0)
                        {
                            IEnumerator<Tile> tileEnumerator = tilesHere.GetEnumerator();
                            tileEnumerator.MoveNext();
                            Tile top = tileEnumerator.Current;
                            game.Profiler.Start("handleCollisions(Tiles,Mob,Tile)");
                                    
                            if (mob1.touches(top))
                            {
                                foreach (Tile tile in tilesHere)
                                {
                                    mob1.touch(tile);
                                    tile.touch(mob1);
                                }
                            }
                            game.Profiler.Stop("handleCollisions(Tiles,Mob,Tile)");
                        }
                    }
                }
                game.Profiler.Stop("handleCollisions(Tiles,Mob)");
            }
            game.Profiler.Stop("handleCollisions(Tiles)");

            game.Profiler.Start("handleCollisions(Mobs)");

            foreach (Mob mob1 in mobs)
            {
                
                ICollection<Mob> nearbyMobs = MobsInBounds(mob1.Bounds);

                foreach (Mob mob2 in nearbyMobs)
                {
                    if (mob1 == mob2)
                    {
                        continue;
                    }

                    if (mob1.touches(mob2))
                    {
                        mob1.touch(mob2);
                    }
                }
            }
            game.Profiler.Stop("handleCollisions(Mobs)");
        }

        #region ICollider Members


        public ICollection<Mob> MobsInBounds(Rect bounds)
        {
            float w = (float)bounds.Width * .5f + .5f;
            float h = (float)bounds.Height * .5f + .5f;
            Vector2 c = new Vector2((float)(bounds.X+bounds.Width/2), (float)(bounds.Y+bounds.Height/2));
            int minX = (int)(c.X - w);
            int minY = (int)(c.Y - h);
            int maxX = (int)(c.X + w);
            int maxY = (int)(c.Y + h);

            if (minX < minTileX)
                minX = minTileX;
            if (minY < minTileY)
                minY = minTileY;
            if (maxX > maxTileX)
                maxX = maxTileX;
            if (maxY > maxTileY)
                maxY = maxTileY;

            
            HashSet<Mob> nearbyMobs = new HashSet<Mob>();

            for (int x = minX; x <= maxX; x++)
            {
                for (int y = minY; y <= maxY; y++)
                {
                    foreach (Mob nearbyMob in mobsAt(x, y))
                    {
                        nearbyMobs.Add(nearbyMob);
                    }
                }
            }
            return nearbyMobs;
        }

        #endregion
    }
}
