﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Media;
using Microsoft.Xna.Framework;

namespace Cursed.geom
{
    public abstract class GeometryShape : IShape
    {
        Geometry geom;
        Vector2 translation;
        float rotation;
        float scale;
        protected GeometryShape(Geometry geom)
        {
            this.geom = geom;
        }

        public Geometry Geometry
        {
            get { return geom; }
            set { geom = value; }
        }

        public Transform Transform
        {
            get { return geom.Transform; }
            set { geom.Transform = value; }
        }

        public Vector2 Translation
        {
            get { return translation; }
            set
            {
                translation = value;
                setTransform(translation, rotation, scale);
            }
        }

        public float Rotation
        {
            get { return rotation; }
            set
            {
                rotation = value;
                setTransform(translation, rotation, scale);
            }
        }

        public float Scale
        {
            get { return scale; }
            set
            {
                scale = value;
                setTransform(translation, rotation, scale);
            }
        }

        #region IShape Members

        public virtual bool intersects(IShape other)
        {
            if (!(other is GeometryShape))
            {
                throw new NotImplementedException("GeometryShape instances can only intersect other GeometryShape instances");
            }

            if (Bounds.IntersectsWith(other.Bounds))
            {
                Geometry otherGeom = ((GeometryShape)other).Geometry;
                if (Geometry.FillContainsWithDetail(otherGeom) != IntersectionDetail.Empty)
                    return true;
            }
            return false;
        }

        public abstract Vector2 intersectionNormal(IShape other);
        public abstract Vector2 Center { get; }
        public System.Windows.Rect Bounds 
        {
            get { return geom.Bounds; }
        }

        #endregion

        #region IShape Members


        public virtual void setTransform(Vector2 translation, float rotation, float scale)
        {
            this.translation = translation;
            this.rotation = rotation;
            this.scale = scale;

            TransformGroup tg = new TransformGroup();
            if (scale != 1)
                tg.Children.Add(new ScaleTransform(scale, scale));
            if (rotation != 0)
                tg.Children.Add(new RotateTransform(180 * rotation));
            if (translation != Vector2.Zero)
                tg.Children.Add(new TranslateTransform(translation.X, translation.Y));
            geom.Transform = tg;
        }

        #endregion
    }

    public class Rect : GeometryShape
    {

        bool canPushUp = true;
        bool canPushDown = true;
        bool canPushLeft = true;
        bool canPushRight = true;

        public Rect(Vector2 center, float width, float height)
            : base(new RectangleGeometry(new System.Windows.Rect(center.X - width / 2, center.Y - height / 2, width, height)))
        {

        }

        public RectangleGeometry RectangleGeometry
        {
            get { return (RectangleGeometry)Geometry; }
            set { Geometry = value; }
        }

        public bool CanPushUp
        {
            get { return canPushUp; }
            set { canPushUp = value; }
        }

        public bool CanPushDown
        {
            get { return canPushDown; }
            set { canPushDown = value; }
        }

        public bool CanPushRight
        {
            get { return canPushRight; }
            set { canPushRight = value; }
        }

        public bool CanPushLeft
        {
            get { return canPushLeft; }
            set { canPushLeft = value; }
        }
        public float Width
        {
            get { return (float)((RectangleGeometry)Geometry).Rect.Width; }
        }
        public float Height
        {
            get { return (float)((RectangleGeometry)Geometry).Rect.Height; }
        }

        void filterNormal(Vector2 normal)
        {
            if (!CanPushUp && normal.Y > 0)
                normal.Y = 0;
            if (!CanPushDown && normal.Y < 0)
                normal.Y = 0;
            if (!CanPushLeft && normal.X < 0)
                normal.Y = 0;
            if (!CanPushRight && normal.X > 0)
                normal.X = 0;
        }

        public override Vector2 intersectionNormal(IShape other)
        {
            if (!(other is GeometryShape))
            {
                throw new NotImplementedException("GeometryShape instances can only intersect other GeometryShape instances");
            }

            if (other is Circle)
            {
                Circle circle = (Circle)other;
                Vector2 xoff = new Vector2(Width / 2, 0);
                Vector2 yoff = new Vector2(0, Height / 2);
                Vector2[] vertices = { Center - xoff - yoff,
                                         Center - xoff + yoff,
                                         Center + xoff + yoff,
                                         Center + xoff - yoff };

                Vector2 closest = Center;

                for (int i = 0; i < 4; i++)
                {
                    Vector2 v1 = vertices[i];
                    Vector2 v2 = vertices[(i + 1) % 4];
                    Vector2 s = (v2 - v1);
                    Vector2 u = s;
                    u.Normalize();

                    Vector2 p = circle.Center - v1;
                    float l = Vector2.Dot(u, p);

                    Vector2 c = v1 + u * l;
                    if (l < 0)
                        c = v1;
                    if (l > s.Length())
                        c = v2;

                    if ((circle.Center - closest).Length() > (circle.Center - c).Length())
                    {
                        closest = c;
                    }
                }

                Vector2 normal = circle.Center - closest;

                normal *= (normal.Length() - circle.Radius) / normal.Length();
                filterNormal(normal);
                return normal;
            }
            else
            {
                Vector2 offset = other.Center - Center;
                Vector2 normal = Vector2.Zero;
                float overlapX = Math.Abs(offset.X) - .5f * (float)Bounds.Width - .5f * (float)other.Bounds.Width;
                float overlapY = Math.Abs(offset.Y) - .5f * (float)Bounds.Height - .5f * (float)other.Bounds.Height;
                if (overlapX < overlapY)
                {
                    normal.Y = overlapY * Math.Abs(offset.Y) / offset.Y;
                }
                else
                {
                    normal.X = overlapX * Math.Abs(offset.X) / offset.X;
                }
                filterNormal(normal);
                return normal;
            }
        }
        
        public override bool intersects(IShape other)
        {
            if (Bounds.IntersectsWith(other.Bounds))
            {
                if (other is Rect && Rotation == 0 && other.Rotation == 0)
                {
                    return true;
                }
                if (other is Circle)
                {
                    return other.intersects(this);
                }
                return base.intersects(other);
            }
            return false;
        }
        
        public override Vector2 Center
        {
            get {
                System.Windows.Point center = new System.Windows.Point(RectangleGeometry.Rect.X + Width * .5f, RectangleGeometry.Rect.Y + Height * .5f);
                center = Transform.Transform(center);
                return new Vector2((float)center.X, (float)center.Y);
            }
        }
    }

    public class Circle : GeometryShape
    {
        public Circle(Vector2 center, float radius)
            : base(new EllipseGeometry(new System.Windows.Point(center.X, center.Y), radius, radius))
        {

        }

        public override Vector2 intersectionNormal(IShape other)
        {
            if (!(other is GeometryShape))
            {
                throw new NotImplementedException("GeometryShape instances can only intersect other GeometryShape instances");
            }

            if (other is Circle)
            {
                Circle circle = (Circle)other;
                Vector2 normal = circle.Center - Center;
                normal *= (normal.Length() - Radius - circle.Radius) / normal.Length();
                
                return normal;
            }
            else
            {
                Vector2 backwardsNormal = other.intersectionNormal(this);
                return backwardsNormal * -1;
            }
        }

        public override bool intersects(IShape other)
        {
            if (!(other is GeometryShape))
                return base.intersects(other);
            GeometryShape gother = (GeometryShape)other;

            Vector2 c = other.Center;
            Vector2 o = c - Center;
            if (o.LengthSquared() < Radius * Radius)
                return true;

            if (other is Circle)
            {
                Circle circle = (Circle)other;
                float min = Radius + circle.Radius;
                float distance = (circle.Center - Center).LengthSquared();
                return distance <= min * min;
            }
            
            if (!Bounds.IntersectsWith(other.Bounds))
                return false;
            if (other is Rect)
            {
                if (gother.Geometry.FillContains(new System.Windows.Point(Center.X, Center.Y)))
                    return true;
                Vector2 offsetNormal = intersectionNormal(other);
                Vector2 offset = other.Center - Center;
                float d = Vector2.Dot(offsetNormal, offset);
                return Vector2.Dot(offsetNormal, other.Center - Center) <= 0;
                //return offsetNormal.LengthSquared() < (other.Center - Center).LengthSquared();
            }
             
            return base.intersects(other);
        }
          
        public override void setTransform(Vector2 translation, float rotation, float scale)
        {
            base.setTransform(translation, 0, scale);
        }

        public EllipseGeometry EllipseGeometry
        {
            get { return (EllipseGeometry)Geometry; }
            set { Geometry = value; }
        }

        public float Radius
        {
            get { return Scale*(float)EllipseGeometry.RadiusX; }
        }

        public override Vector2 Center
        {
            get {
                System.Windows.Point center = new System.Windows.Point(EllipseGeometry.Center.X, EllipseGeometry.Center.Y);
                center = Transform.Transform(center);
                return new Vector2((float)center.X, (float)center.Y);
            }
        }
    }
}
