﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using Cursed.view;
using System.Windows.Media;
using Cursed.input;
using Cursed.gob.tile;
using Cursed.geom;
using System.Runtime.Serialization;
using Cursed.view.renders;


namespace Cursed.gob.mob
{
    public class MobInfo : GobInfo
    {
        float mass; //for collisions, perhaps accels. applying a force would consider mass.
        bool stuck;

        Vector2 vel;
        float accMag;

        float damping;
        float maxSpeedSquared;

        float theta; //rotation
        float omega; //angular velocity

        float scale;

        public MobInfo()
            : base()
        {
            mass = 1;
            stuck = false;
            vel = Vector2.Zero;
            accMag = 1;
            damping = 0;
            maxSpeedSquared = 100;
            theta = 0;
            omega = 0;
            scale = 1;
        }

        public float Mass
        {
            get { return mass; }
            set { mass = value; }
        }
        public bool Stuck
        {
            get { return stuck; }
            set { stuck = value; }
        }
        public Vector2 Vel
        {
            get { return vel; }
            set { vel = value; }
        }
        public float AccMag
        {
            get { return accMag; }
            set { accMag = value; }
        }
        public float Damping
        {
            get { return damping;}
            set { damping = value;}
        }
        public float MaxSpeedSquared
        {
            get { return maxSpeedSquared; }
            set { maxSpeedSquared = value; }
        }
        public float Theta
        {
            get { return theta; }
            set { theta = value; }
        }

        public float Omega
        {
            get { return omega; }
            set { omega = value; }
        }
        public float Scale
        {
            get { return scale; }
            set { scale = value; }
        }
    }


    public class Mob : Gob, IControllable, IFocus
    {
        
        
        public const float mobPushOffFactor = 0.002f;

        //direction the control is telling this mob to move at the moment
        protected Vector2 direction;
        //this time step's accel vector
        protected Vector2 acc;

        protected RenderInfo renderInfo;

        public Mob(CursedGame game, MobInfo info)
            : base(game, info)
        {

        }


        public override void Update(GameTime gameTime)
        {
            float seconds = gameTime.ElapsedGameTime.Milliseconds*.001f;

            calcPos(seconds);

            if (renderInfo != null)
            {
                updateRenderTransform(renderInfo);
            }

            updateTransform();

            base.Update(gameTime);
        }
        public MobInfo MobInfo
        {
            get { return (MobInfo)info; }
        }

        public void ApplyForce(Vector2 f)
        {
            acc += f * (1 / MobInfo.Mass);
        }

        public void doStopCollision(Gob gob)
        {
            Vector2 refVelocity = Vector2.Zero;
            Vector2 relVelocity = MobInfo.Vel;
            if (gob is Mob)
            {
                refVelocity = ((Mob)gob).MobInfo.Vel;
            }
            relVelocity -= refVelocity;

            Vector2 normal = Shape.intersectionNormal(gob.Shape);
            Vector2 unit = normal;
            unit.Normalize();

            float relInlineSpeed = Vector2.Dot(normal, relVelocity);

            if (relInlineSpeed < 0)
            {
                Vector2 inVel = unit * relInlineSpeed;
                Vector2 offVel = relVelocity - inVel;
                MobInfo.Vel = refVelocity + offVel;
            }

            pushOff(gob, normal);
        }

        /*
         * Push self off of gob in small increments, until no longer touching
         */
        protected void pushOff(Gob gob, Vector2 normal)
        {
            if (normal == Vector2.Zero)
                return;// normal = new Vector2((float)CursedGame.Random.NextDouble(), (float)CursedGame.Random.NextDouble());
            
            Vector2 pushDistance = normal;
            //if (pushDistance.Length() > Const.maxPush)
            //    pushDistance *= Const.maxPush / pushDistance.Length();
            float ratio = 0;
            if (gob is Mob && !((Mob)gob).MobInfo.Stuck)
            {
                ratio = MobInfo.Mass/(MobInfo.Mass + ((Mob)gob).MobInfo.Mass);
                pushDistance = normal * .5f;
                gob.GobInfo.Loc -= normal * ratio;
                ((Mob)gob).updateTransform();
            }
            GobInfo.Loc += normal * (1 - ratio);
            updateTransform();
            
            /*
            bool isMe = true;
            Vector2 push = normal;// *mobPushOffFactor / normal.Length();
            
            do
            {
                Mob m = this;
                if (!isMe)
                {
                    m = (Mob)gob;
                }

                m.GobInfo.Loc -= push;
                m.updateTransform();

                if (gob is Mob)
                {
                    isMe = !isMe;
                    push *= -1;
                }
            } while (touches(gob));
            */
        }

        protected void updateTransform()
        {
            shape.setTransform(MobInfo.Loc, MobInfo.Theta, MobInfo.Scale);
        }

        protected void updateRenderTransform(RenderInfo rInfo)
        {
            rInfo.Transform = Microsoft.Xna.Framework.Matrix.CreateScale(MobInfo.Scale);
            rInfo.Transform *= Microsoft.Xna.Framework.Matrix.CreateRotationZ(MobInfo.Theta);
            rInfo.Transform *= Microsoft.Xna.Framework.Matrix.CreateTranslation(MobInfo.Loc3);
        }

        private void calcPos(float seconds)
        {

            float friction = 0;
            float drag = 0;

            foreach (Gob touchedGob in touchedGobs)
            {
                if (touchedGob is Ground)
                {
                    Ground ground = (Ground)touchedGob;
                    if (ground.GroundInfo.Drag > drag)
                        drag = ground.GroundInfo.Drag;
                    if (ground.GroundInfo.Friction > friction)
                        friction = ground.GroundInfo.Friction;
                }
                if (!MobInfo.Stuck && !Transcendent && !touchedGob.Transcendent)
                {
                    doStopCollision(touchedGob);
                }
            }

            acc += direction * MobInfo.AccMag * friction;
            acc -= drag * MobInfo.Vel;
            //accel = a - kv
            acc -= MobInfo.Damping * MobInfo.Vel;

            //v = v0 + at
            MobInfo.Vel += acc * seconds;

            //pos = p0 + v*t + 1/2 a*t^2;
            MobInfo.Loc += MobInfo.Vel * seconds + acc*.5f*seconds*seconds;

            acc = Vector2.Zero;
        }

        public virtual void control(ControlInfo cInfo)
        {
            if (cInfo is DirectionControlInfo)
            {
                direction = ((DirectionControlInfo)cInfo).Direction;
            }
        }
        
        #region IFocus Members

        public Vector2 Loc
        {
            get { return MobInfo.Loc; }
        }

        #endregion
    }
}
